//  $Id: pfeap.hh,v 1.8 2004/08/20 07:19:02 adams Exp $
//  Author:             Mark F. Adams
// Copyright (c) 2003 by Mark F. Adams 
//  Filename:           pfeap.hh
//
#ifndef  __PFEAP_H__
#define  __PFEAP_H__

#include "prometheus.hh"

class Prometheus_LinSysCore;

////////////////////////////////////////////////////////////////////////
//! Interface to FE application
////////////////////////////////////////////////////////////////////////
class FEComponentIterface
{
public:
  virtual int GetSolution( PromVector *const XX ) = 0;
  virtual int GetResidual( PromVector *const RR ) = 0;
  virtual int ComputeResidual() = 0;
  virtual int Reac( int nnodes ) = 0;
  virtual int FormTangent(Prometheus_LinSysCore*const)=0;
  virtual int FormTangentAndResidual(Prometheus_LinSysCore*const) = 0;
  virtual int ReadMesh(char *feapname, int nlocalelems=0)=0;
  virtual int SetDT( double ) = 0;
  virtual int StepTime() = 0;
  virtual int Finalize() = 0;
};

////////////////////////////////////////////////////////////////////////
//! Interface for data output
////////////////////////////////////////////////////////////////////////
class FEDataClient
{
public:
  virtual int GetMeshSpecs(int*nNds,int*nElems,int*nLocEls,int*ndf)const=0;
  virtual int GetNElemBlks(int *nsets)const=0;
  virtual int GetNMaterials( int *nm )const=0;
  virtual int GetNElemsDegree(int ElBlkID, int *nconn)const=0;
  virtual int GetNNodes( int * )const = 0;
  virtual int GetElemMeshBlk(int ElBlkID, double coords[], 
			     int matids[], int elems[] )const=0;
  virtual int GetElemBlkAtr(int ElBlkID, int nel, float* )const=0;
  virtual int GetNElemAtr(int ElBlkID, int *na )const=0;
  virtual int GetNodeDisp( float array[], int)const=0;
};

////////////////////////////////////////////////////////////////////////
//! Interface to FE application -- FEAP
////////////////////////////////////////////////////////////////////////
class pFEAP : public FEComponentIterface, public FEDataClient
{
public:
  pFEAP(const PromComm &comm, PromOptions &opts, PromPerfMonitor &perf);
  virtual ~pFEAP();
  // FEDataClient
  int GetMeshSpecs( int *nNds, int *nElems, int *nLocEls, int * )const;
  int GetNElemBlks(int *nsets)const{ *nsets=1; return 0;}
  int GetNMaterials( int *nm )const;
  int GetNElemsDegree(int ElBlkID, int *nconn)const;
  int GetNNodes( int *nn )const;
  int GetElemMeshBlk( int ElBlkID, double coords[], 
		      int matids[], int elems[] )const;
  int GetElemBlkAtr(int ElBlkID, int nel, float * )const;
  int GetNElemAtr(int ElBlkID, int *na )const;
  int GetNodeDisp( float array[], int ) const;
  // FEComponentIterface
  int GetSolution( PromVector *const XX );
  int GetResidual( PromVector *const RR );
  int ComputeResidual();
  int Reac( int nnodes );
  int FormTangent(Prometheus_LinSysCore*const);
  int FormTangentAndResidual(Prometheus_LinSysCore*const);
  //
  //int CopyVectorOut( PromVector *const BB, double const*const array );
  //int CopyVectorIn( const PromVector *const XX, double *const array );
  // methods
  int ReadMesh( char *feapname, int nlocalelems = 0 );
  int SetUpLinSys(Prometheus_LinSysCore *prom, const int npLoc, 
		  const int npGst, const int ghost_gnode[], int &my0Nd );
  int SetDT( double );
  int StepTime();
  int Finalize();
  // feap displacemetn output
  int WriteDispFile() const;
  int WriteDisp( const int step, int )const;
  // util
  int SetMaps( const int nlocalnd, const int pnode_fnode[],
	       const int ndf, const int my0Nd );
private:
  int GetSolution_private( const PromVector *const XX );
  // data
protected:
  IS    lpetsc_gfeapIS_;  // used to put petsc.x back to real feap
  IS    lfeap_gpetscIS_;  // used to put petsc.b out of  feap
  IS    afeap_gpetscIS_;  // used to put petsc.x into feap
public:
  int   nfixed_;          // includes ghost nodes
  int   nLocalFixed_;     // includes non-ghost nodes only
  int   nNodes_;
  int   nneq_;            // total equations locally
  int   feapEQIDsz_;      // feapEQIDsz_ = nneq - nLocalFixed_
  int   nLocElems_;
  int  *promeq_appeq_; //[feapEQIDsz_]: X = promeq_appeq_[promEq] + nfixed_
  int  *feapEQID_lid_; //[feapEQIDsz_+1] lid(X) = feapEQID_lid_[X]
  int  *lid_gid_;
  // data
  const PromComm &Comm_; 
  PromOptions &options_;
  PromPerfMonitor &perf_mon_;
protected:
  Vec  work_nfl_;
  Vec  work_neq_; 
  IS   identID_nfl_;
  IS   identID_neq_; 
  VecScatter scatter_nfl_;
  VecScatter scatter_neq_;
};

#endif //  __PFEAP_H__
