//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_grid_mis.C,v 1.8 2004/11/05 23:03:23 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prom_grid_mis.C
//
#include "prom_grid.hh"

#undef __FUNCT__    
#define __FUNCT__ "PromMatrix::GetDeps" 
/* PromMatrix::GetDeps ***********************************************
 *  
 *  INPUT:
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *    - PETSc cerror code 
 */
int PromMatrix::GetDeps( int gidi, int &ndeps, int procs[] )

{
  int proc,ierr,ncols,ii,jj,*adjacs,gidj, np = map_.comm_.np();
  const int *p_g=map_.proc_gnode(),my0=map_.my0Nd(),next0=my0+map_.nlocal();
  
  ierr = GetRow( gidi, &ncols, &adjacs); CHKERRQ(ierr);
  for( ii = ndeps = 0 ; ii < ncols ; ii++ ) {
    gidj = adjacs[ii];
    if( gidj < my0 || gidj >= next0  ) {
      ierr = PromGrid::FindProc( p_g, np, gidj, proc ); CHKERRQ(ierr);
      for( jj = 0 ; jj < ndeps ; jj++ ) {
	if( procs[jj] == proc ){ gidj = -1; break; } 
      }
      if( gidj != -1 ) procs[ndeps++] = proc; 
    }
  }
  ierr = RestoreRow( gidi, &ncols, &adjacs ); CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromMatrix::CreateMIS" 
/* PromMatrix::CreateMIS ***********************************************
 *  
 *  - take this mesh and coarsen, to make (almost) max. ind. sets 
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *    - PETSc cerror code 
 */
int PromMatrix::CreateMIS( PromGrid *const grid,
			   PromList<PromNode*> &nodeList,
			   int Isdeleteable(const int,const int,bool &))
{
  PromTable<int> procTable(57), procTable_orig(57); bool flag;
  int ierr,ii,xx,receive,doit,proc,*pb,ndone,its,todo,todo2,tt,*garr;
  int nghosts,gidi,gidj,ndeps,procs[1024];
  const int nloc=map_.nlocal(),*p_g=map_.proc_gnode();
  const int mype=map_.comm_.mype(),np=map_.comm_.np();
  const int tag=grid->getNewTag(),tag1=grid->getNewTag();
  MPI_Status status;  TablePos tpos;
  Position pos,pos2;  PromNode *curr, *n1;
  TablePos tipos; MPI_Comm gcomm = grid->gComm_; assert(grid->isActive());
  const int nnodes_orig = grid->nNodes_,gtag=grid->getNewTag(),tag2=grid->getNewTag();
  const int reqsz=sizeof(MPI_Request)/sizeof(int);
  PromList<PromNode*> *list; NdStTAG state;
  PromList<MPI_Request*> smessages; 
  PromTable<PromList<PromNode*>*> proc_list(57);
  PromTable<NdStTAG> gid_state( 2*(nnodes_orig-nloc) + 5);

  PetscFunctionBegin;  
  // clear, make gid_ghost and procTable
  ierr = GetGArray( &garr, &nghosts ); CHKERRQ(ierr); 
  for( xx = 0 ; xx < nnodes_orig ; xx++ ) {
    curr = grid->nodes_[xx]; curr->mark(FALSE);
    gidi = curr->index_.global();
    if( xx < nloc ) {
      gid_state.add( PROM_UNKNOWN, gidi+1 );
      ierr = GetDeps( gidi, ndeps, procs ); CHKERRQ(ierr); 
      for( ii = 0 ; ii < ndeps ; ii++ ) {
	proc = procs[ii];
	if( !proc_list.find(proc+1) ){
	  list = new PromList<PromNode*>(100); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	  procTable_orig.add( proc + 1 );
	}
      }
    }
  }
  for( xx = 0 ; xx < nghosts ; xx++ ) {
    gidj = garr[xx];
    gid_state.add( PROM_UNKNOWN, gidj+1 );
    ierr = PromGrid::FindProc( p_g, np, gidj, proc ); CHKERRQ(ierr);
    if( !proc_list.find(proc+1) ){
      list = new PromList<PromNode*>(100); CHKNEW(list,1);
      proc_list.add( list, proc+1 );
      procTable_orig.add( proc + 1 );
    }
  }
  ierr = RestoreGArray( &garr ); CHKERRQ(ierr); 
  todo2 = nloc;
  assert(todo2 == nodeList.getCount()); 

  // copy in
  tipos = procTable_orig.getHeadPosition(); 
  while( tipos ) procTable.add( procTable_orig.getNext(tipos) );
  // MIS
  int ptodo = procTable_orig.getCount();
  for( its = todo = 0 ; /* VOID */ ; its++ ){
    // one MIS iteration
    pos = nodeList.getHeadPosition(); ndone = 0;
    while( (pos2=pos) ){
      curr = nodeList.getNext(pos);  gidi = curr->index_.global();
      if( (state=gid_state.find( gidi+1 )) != PROM_UNKNOWN ){ 
	nodeList.removeAt(pos2);
	// send DELETED bound
	if( state == PROM_DELETED && !curr->isMarked() ) {
	  curr->mark(); n1 = curr->symbData_->parent; assert(n1!=NULL);
	  gidj = n1->index_.global(); 
#if defined(PETSC_USE_BOPT_g)
	  assert(gid_state.find(gidj+1)==PROM_SELECTED);
#endif
	  ierr = GetDeps( gidi, ndeps, procs ); CHKERRQ(ierr); 
	  for( ii = 0 ; ii < ndeps ; ii++ ) {
	    proc = procs[ii];	    // send to deps
	    list = proc_list.find(proc+1);    assert(list!=NULL);
	    list->addTail( curr ); list->addTail( n1 ); 
	  }
	}
      }
      else {
	assert(state!=0);
	// check for selectable
	doit = TRUE; //(!curr->isAllFixed() || its>0); // ***
	int nn,adj_arr[PROM_MAXNIEGH];
	{ int ncols,*adjacs;
	ierr = GetRow( gidi, &ncols, &adjacs ); CHKERRQ(ierr);
	for(ii=nn=0;ii<ncols;ii++)if(gidi!=adjacs[ii]) adj_arr[nn++]=adjacs[ii];
	assert(nn==ncols-1);
	ierr = RestoreRow( gidi, &ncols, &adjacs ); CHKERRQ(ierr);
	}
	//
	for( ii = 0 ; ii < nn ; ii++ ) {
	  gidj = adj_arr[ii]; 
	  if( (state = gid_state.find(gidj+1)) == PROM_UNKNOWN ) {
	    ierr = PromGrid::FindProc( p_g, np, gidj, proc ); 
	    CHKERRQ(ierr);
	    if( proc > mype ) { 
	      ierr = Isdeleteable( gidj, gidi, flag ); CHKERRQ(ierr); 
	      if( flag ) doit = FALSE;
	    }
	  }
	  else if( state == PROM_SELECTED ) {
	    ierr = Isdeleteable( gidj, gidi, flag ); CHKERRQ(ierr); 
	    if( flag ) doit = FALSE; 
	  }
	}

	if( doit ) {
	  // select
	  gid_state.add( PROM_SELECTED, gidi+1 );  todo2--;
	  for( ii = 0 ; ii < nn ; ii++ ) {
	    gidj = adj_arr[ii]; 
 	    if( (state = gid_state.find(gidj+1)) == PROM_UNKNOWN ) {
	      ierr = Isdeleteable( gidi, gidj, flag ); CHKERRQ(ierr);
	      if( flag ) { 
		ierr = PromGrid::FindProc(p_g, np, gidj, proc); CHKERRQ(ierr);
		if( proc != mype ){
		  list = proc_list.find(proc+1); assert(list!=NULL);
		  todo += grid->addGhost( n1, proc, gidj, 
					  gidj - p_g[proc], gtag );
		  list->addTail( n1 ); list->addTail( curr );
		}
		else{ todo2--; n1 = grid->nodes_[gidj-p_g[mype]]; }
		gid_state.add( PROM_DELETED, gidj+1 );
		n1->symbData_->parent = curr;
		ndone++;
	      }
	    }
	  }
	}
      }
    }

    // isend(tag)
    tpos = proc_list.getHeadPosition();
    while( tpos ){ 
      list = proc_list.getNext( tpos, &proc ); proc--;
      if( list->isEmpty() ) continue; // do not send NUL as is capper flag
      tt = list->getCount();        assert(tt%2==0 && tt>0);
      if( tt ) {
	MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
	CHKERRQ(ierr);	smessages.addTail( sreq );
	int *buff = (int*)(sreq+1), *pb = buff;
	while( !list->isEmpty() ) { 
	  n1 = list->removeHead(); *pb++ = n1->index_.global();
	}
	assert(tt == pb-buff);
	// send(tag) *<gid_D_gh, gid_S_loc> | <gid_loc_DEL, pargid_SEL>
	MPI_Isend( buff, tt, MPI_INT, proc/grid->factor_, tag, gcomm, sreq);
      }
    }
    // loop test
    if ( todo2 == 0 ) break;

    // receive(tag)
    do{
      tpos = proc_list.getHeadPosition();  
      while( tpos ) { 
	proc_list.getNext( tpos, &proc ); proc--;
	do{
	  MPI_Iprobe( proc/grid->factor_, tag, gcomm, &receive, &status );
	  if( receive ) {
	    MPI_Get_count( &status, MPI_INT, &tt ); 
	    if( tt == 0 ) {
	      MPI_Recv(NULL,0,MPI_INT,status.MPI_SOURCE,tag,gcomm,&status);
	      ptodo--;
	    }
	    else {
	      // recv(tag) *< gid_D_loc, gid_S_gh> | < gid_gh_D, pargid_S>
	      ierr = ReceiveState_NC(grid, grid->ghosts_, gid_state, tt, 
				     ndone, todo2, todo, tag, gtag, 
				     status.MPI_SOURCE, gcomm );
	      CHKERRQ(ierr);
	    }
	  }
	}while( receive && todo2 );
      }
    }while( ndone == 0 && todo2 ); // wait till something to do
  } // main MIS loop

  // collect PROM_DELETED bound
  for( xx = 0 ; xx < nloc ; xx++ ) {
    curr = grid->nodes_[xx]; gidi = curr->index_.global();
    if( gid_state.find(gidi+1) == PROM_DELETED && !curr->isMarked() ) {
      curr->mark(); n1 = curr->symbData_->parent;
#if defined(PETSC_USE_BOPT_g)
      assert(n1!=NULL && gid_state.find(n1->index_.global()+1)==PROM_SELECTED);
#endif
      ierr = GetDeps( gidi, ndeps, procs ); CHKERRQ(ierr); 
      for( ii = 0 ; ii < ndeps ; ii++ ) {
	proc = procs[ii];
	list = proc_list.find(proc+1); assert(list!=NULL);
	list->addTail( curr ); list->addTail( n1 ); 
      }
    }
  }

  // send PROM_DELETED bound
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc); proc--;
    tt = list->getCount(); 
    if( tt > 0 ) {
      MPI_Request *sreq; 
      ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq); CHKERRQ(ierr);
      smessages.addTail( sreq );
      int *buff = (int*)(sreq+1), *pb = buff;
      while( !list->isEmpty() ) {
	curr = list->removeHead(); 
	*pb++ = curr->index_.global();
      }
      assert(tt == pb-buff);
      // send(tag) *<gid_D_loc, gid_S_gh> | <gid_gh_D, pargid_S>
      MPI_Isend(buff, tt, MPI_INT, proc/grid->factor_, tag, gcomm,sreq);
    }
    delete list; 
  }
  // send(tag) capper
  tipos = procTable_orig.getHeadPosition();
  while( tipos ) {
    proc = procTable_orig.getNext(tipos) - 1;
    MPI_Send( NULL, 0, MPI_INT, proc/grid->factor_, tag, gcomm );
  }

  // receive(tag) and capper(tag) 
  do{
    MPI_Iprobe( MPI_ANY_SOURCE, tag, gcomm, &receive, &status );
    if( receive ) {
      MPI_Get_count( &status, MPI_INT, &tt ); 
      if( tt == 0 ) {
	MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE,tag,gcomm,&status);
	ptodo--; assert(ptodo>=0);
      }
      else {
	// recv(tag) *< gid_D_loc, gid_S_gh> | < gid_gh_D, pargid_S >
	ierr = ReceiveState_NC(grid, grid->ghosts_, gid_state, tt, 
			       ndone, todo2, todo,tag,gtag, 
			       status.MPI_SOURCE, gcomm );
	CHKERRQ(ierr);
      }
    }
  }while( ptodo );
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }

  // collect send(tag2) PROM_DELETED bound
  assert( proc_list.isEmpty() );
  for( xx = 0 ; xx < nloc ; xx++ ) {
    curr = grid->nodes_[xx]; gidi = curr->index_.global();
    if( gid_state.find(gidi+1) == PROM_DELETED && !curr->isMarked() ) {
      curr->mark(); n1 = curr->symbData_->parent;
#if defined(PETSC_USE_BOPT_g)
      assert(n1!=NULL && gid_state.find(n1->index_.global()+1)==PROM_SELECTED);
#endif
      ierr = GetDeps( gidi, ndeps, procs ); CHKERRQ(ierr); 
      for( ii = 0 ; ii < ndeps ; ii++ ) {
	proc = procs[ii];
	if( (list=proc_list.find(proc+1)) == NULL ){
	  list = new PromList<PromNode*>(10); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	list->addTail( curr );
      }
    }
  }
  
  // send(tag2) PROM_DELETED bound
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc); proc--;
    tt = 2*list->getCount(); 
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);	smessages.addTail( sreq );
    int *buff = (int*)(sreq+1), *pb = buff;
    while( !list->isEmpty() ) {
      curr = list->removeHead();
      n1 = curr->symbData_->parent;
      *pb++ = curr->index_.global(); *pb++ = n1->index_.global();
    }
    assert(tt == pb-buff);
    // send(tag2) *< gid_loc_DEL, pargid_SEL >
    MPI_Isend( buff, tt, MPI_INT, proc/grid->factor_, tag2, gcomm, sreq );
    delete list;
  }

  // capper (tag2)
  tipos = procTable_orig.getHeadPosition(); ptodo = 0;
  while( tipos ) {
    proc = procTable_orig.getNext(tipos) - 1; ptodo++;     
    MPI_Send(NULL, 0, MPI_INT, proc/grid->factor_, tag2, gcomm );
  }
  while( ptodo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag2, gcomm, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    if( tt == 0 ) {
      ptodo--;
      MPI_Recv(NULL,0,MPI_INT,status.MPI_SOURCE,tag2,gcomm,&status);
    }
    else {
      // recv(tag2) *< gid_loc_DEL, pargid_SEL >
      ierr = ReceiveState_NC( grid, grid->ghosts_, gid_state, tt, ndone,
			      todo2, todo, tag2, gtag, 
			      status.MPI_SOURCE, gcomm );
      CHKERRQ(ierr);
    } 
  }
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  
  // set extras selected (will get done in add...)
  for( xx = nloc ; xx < nnodes_orig ; xx++ ) {
    curr = grid->nodes_[xx]; gidi = curr->index_.global();
    if( gid_state.find(gidi+1) == PROM_UNKNOWN ){
      gid_state.add( PROM_SELECTED, gidi+1 );
      //curr->setState(PROM_SELECTED);
    }
  }

  { // get all procs for 'ReceiveAddGhost'
    int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 
    pb = sbuf;
    tipos = procTable_orig.getHeadPosition(); 
    while( tipos ) *pb++ = procTable_orig.getNext(tipos) - 1; 
    int nsen = procTable_orig.getCount(); 
    MPI_Request *sreqs = NULL; MPI_Status *statuses = NULL;
    if( nsen ) {
      ierr = PetscMalloc( sizeof(MPI_Request)*nsen, &sreqs );    CHKERRQ(ierr);
      ierr = PetscMalloc( sizeof(MPI_Status)*nsen, &statuses );  CHKERRQ(ierr);
    }
    // send(tag1) <orig procs>
    for( tipos = procTable_orig.getHeadPosition(), ii = 0 ; tipos ; ii++ ) {
      proc = procTable_orig.getNext(tipos) - 1;     
      MPI_Isend(sbuf,pb-sbuf,MPI_INT, proc/grid->factor_,tag1,gcomm,&sreqs[ii]);
    }
    // receive(tag1)
    tipos = procTable_orig.getHeadPosition(); 
    while( tipos ) {
      procTable_orig.getNext(tipos); // counter only
      MPI_Recv( rbuf, PROM_BSZ, MPI_INT, MPI_ANY_SOURCE, tag1,gcomm,&status);
      MPI_Get_count( &status, MPI_INT, &tt );
      for ( pb = rbuf ; tt ; tt-- ){
	proc = *pb++; if( proc != mype ) procTable.add( proc+1 );
      }
    }
    if( nsen ) {
      MPI_Waitall(nsen,sreqs,statuses); PetscFree(sreqs); PetscFree(statuses);
    }
  }
  if(grid->options_.verbose_>1){
    PetscPrintf(gcomm,"[%d]%s %d new ghosts(1)\n",mype,__FUNCT__,todo);
  }

  // set states
  xx = map_.my0Nd(); 
  tpos = gid_state.getHeadPosition();
  while( tpos ){ 
    state = gid_state.getNext( tpos, &gidi ); gidi--;
    curr = grid->ghosts_.find( gidi+1 );
    if( curr == NULL ) curr = grid->nodes_[gidi-xx];
    if( curr->getState() != PROM_UNKNOWN ) SETERRQ(1,"st==PROM_UNKNOWN");
    assert(state!=PROM_UNKNOWN);
    curr->setState(state);
  }

  // set data structures
  ierr = grid->ReceiveAddGhost( todo, gtag, procTable ); CHKERRQ(ierr);

  // clean up and check
  for( xx=0 ; xx < grid->nNodes_ ; xx++) grid->nodes_[xx]->mark(FALSE);
  MPI_Iprobe(MPI_ANY_SOURCE, tag, gcomm, &ii, &status);assert(!ii);
  MPI_Iprobe(MPI_ANY_SOURCE, tag1, gcomm, &ii,&status);assert(!ii);
  MPI_Iprobe(MPI_ANY_SOURCE, tag2, gcomm, &ii,&status);assert(!ii);
  MPI_Iprobe(MPI_ANY_SOURCE, gtag, gcomm, &ii,&status);assert(!ii);

  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::ReceiveState_NC" 
/* PromMatrix::ReceiveState_NC ********************************************
 *
 *  INPUT:
 *   - grid:
 *   - gid_ghost:
 *   - gid_state:
 *   - sz: 
 *   - ndone: (in/out)
 *   - todo2: (in/out)
 *   - todo: (in/out)
 *   - tag: this and proc could just be the status ...
 *   - part:
 *   - comm: 
 *
 *  SIDE EFFECTS:
 *  RETURN: 
 *     - PETSc error code 
 */
int PromMatrix::ReceiveState_NC( PromGrid *const grid, 
				 PromTable<PromNode*> &gid_ghost,
				 PromTable<NdStTAG> &gid_state, 
				 const int sz, int &ndone, int &todo2,
				 int &gtodo, const int rtag, 
				 const int stag, const int part, 
				 MPI_Comm comm )
{
  int ii, gid_del, gid_sel, proc2, ierr, *buff, *pb;
  const int mype=map_.comm_.mype(),nloc=map_.nlocal();
  const int *p_g=map_.proc_gnode(),np=map_.comm_.np();;
  MPI_Status  status;
  PromNode  *select,*del; 
  int rbuf[PROM_BSZ]; 

  if( sz <= PROM_BSZ ) buff = rbuf;
  else{ ierr = PetscMalloc(sz*sizeof(int),&buff); CHKERRQ(ierr); }
  pb = buff; assert(sz%2==0 && sz>0); 
  
  // recv(rtag) < gid_del_D_loc, gid_del_S_gh > or < gid_del_D_gh, par_S >
  MPI_Recv( buff, sz, MPI_INT, part, rtag, comm, &status ); 
  while( pb-buff < sz ) {
    gid_del = *pb++;  gid_sel = *pb++; 
    del = gid_ghost.find( gid_del+1 ); 
    if( del == NULL ) { // deleted local
      // recv(rtag) < gid_del_DEL_loc, gid_sel_SEL_gh >
      select = gid_ghost.find( gid_sel+1 ); assert(select!=NULL);
      // gid_PROM_DELETED_loc
      ii = gid_del - p_g[mype]; assert(ii>=0 && ii < nloc);
      del = grid->nodes_[ii];
#if defined(PETSC_USE_BOPT_g)
      assert(gid_state.find(gid_del+1)!=PROM_SELECTED);
#endif
      if( gid_state.find(gid_sel+1) == PROM_UNKNOWN ) {
	gid_state.add( PROM_SELECTED, gid_sel+1 ); ndone++; 
	//select->setState(PROM_SELECTED);        	
      }
      if( gid_state.find(gid_del+1) == PROM_UNKNOWN ){
	//del->setState(PROM_DELETED); 
	gid_state.add( PROM_DELETED, gid_del+1 );
	del->symbData_->parent = select; todo2--; 
      } 
      else {
#if defined(PETSC_USE_BOPT_g)
	assert(gid_state.find(gid_del+1)==PROM_DELETED && del->symbData_->parent != NULL);
#endif
	if( !del->symbData_->parent->isGhost() ) proc2 = mype;
	else proc2 = del->symbData_->parent->symbData_->ghostID.proc();
	if( select->symbData_->ghostID.proc() > proc2 ){
	  del->symbData_->parent = select;  // PROM_DELETED
	  del->mark(FALSE); // send again as changed
	}
      }
    }
    else { // delete ghost
      // recv(rtag) select < gid_del_D_gh, pargid_S > 
      ii = gid_sel - p_g[mype]; 
      if( ii >= 0 && ii < nloc ){ select = grid->nodes_[ii]; proc2 = mype; }
      else {
	ierr = PromGrid::FindProc(p_g, np, gid_sel, proc2); CHKERRQ(ierr);
	//ierr = PromGrid::GetProc( gid_sel, proc2 );  CHKERRQ(ierr); 
	assert(proc2!=mype);
	gtodo += grid->addGhost(select, proc2, gid_sel, gid_sel-p_g[proc2], stag);
      }
#if defined(PETSC_USE_BOPT_g)
      assert( gid_state.find(gid_sel+1) != PROM_DELETED);
#endif
      //select->setState( PROM_SELECTED );
      gid_state.add( PROM_SELECTED, gid_sel+1 );
      if( gid_state.find(gid_del+1) == PROM_UNKNOWN ) {  
	gid_state.add( PROM_DELETED, gid_del+1 );
	//del->setState(PROM_DELETED); 
	del->symbData_->parent = select;  ndone++;
      }
      else{
#if defined(PETSC_USE_BOPT_g)
	assert(gid_state.find(gid_del+1) == PROM_DELETED);
#endif
	assert(del->symbData_->parent != NULL);
	int pp = del->symbData_->parent->isGhost() ?
	  del->symbData_->parent->symbData_->ghostID.proc() : mype;
	if( proc2 == mype ) assert(del->symbData_->parent == select);
	if( proc2 > pp ) del->symbData_->parent = select;  // DELETE -- change
      }
    }
  }
  assert(pb-buff==sz);
  if( buff != rbuf )  PetscFree(buff);
  
  return 0;
}

/* PromGrid::ReadMISFile() **********************************
 *
 *  - helper
 *
 *  INPUT:
 *   - file:
 *   - flag: all is well (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - 
 */
#undef __FUNCT__  
#define __FUNCT__ "PromGrid::ReadMISFile" 
int PromGrid::ReadMISFile( FILE *file, int &flag /*out*/)
{ 
  int            xx;
  const int mype = comm_.mype();
  PromNode       *curr;
  // make lookup table
  PromTable<int> fid1_lid1( nNodes_ + 1 );
 
  for( xx = 0 ; xx < nNodes_ ; xx++ ){ 
    curr = nodes_[xx];  
    //int kk = curr->index_.feap(ierr)+1; // hack for this failure
    int kk = curr->index_.global()+1; // not used!!!
    //if( kk <= 0 ){ flag = 0; fid1_lid1.removeAll(); return 0; }
    kk = fid1_lid1.add( curr->index_() + 1, kk ); assert(kk);
  } 
  // read file - set states
  int jj = 0, st, fid; 
  for( xx = 0 ; xx < nglobal() ; xx++ ){ 
    // read fid, state
    int ii = fscanf( file, " %d %d", &fid, &st ); 
    if( ii != 2 ){
      PetscPrintf(gComm_,"[%d]%s Abort Reading MIS\n",mype,__FUNCT__);
      for( xx = 0 ; xx < nNodes_; xx++ ){ 
	curr = nodes_[xx]; 
	curr->setState(PROM_UNKNOWN); assert(curr->symbData_->parent == NULL);
      }
      break;
    }
    int lid1 = fid1_lid1.find(fid+1);
    if( lid1 ){ 
      jj++; lid1--;
      assert( lid1 < nNodes_ ); nodes_[lid1]->setState( (PromNodeState)st ); 
    }
  }
  fclose( file ); 	
  if( jj != nNodes_ ) flag = 0; // file no good
  else {
    // set symbData_->parents
    for( xx = 0 ; xx < nNodes_  ; xx++ ){ 
      curr = nodes_[xx];
      if( curr->getState() == PROM_SELECTED ) {
	TablePos tpos = curr->symbData_->adjac.getHeadPosition(); 
	while( tpos ) {  
	  PromNode *n1 = curr->symbData_->adjac.getNext(tpos);
	  if(!n1->symbData_->parent && n1->getState() == PROM_DELETED){
	    n1->symbData_->parent = curr;
	  }
	}
      }
    }
    // check par
    for( xx = 0 ; xx < nNodes_ ; xx++ ){ 
      curr = nodes_[xx];
      if( curr->getState() == PROM_DELETED && !curr->symbData_->parent ) {
	SETERRQ(1,"curr->getState() == DELETED && !curr->symbData_->parent");
	//ierr = curr->SetSymbData_->Parent( 4 );             CHKERRQ(ierr);
      }
      assert( curr->symbData_->parent || curr->getState() == PROM_SELECTED );
    }
  }
  return 0;
}

/* PromGrid::SendStatesToSlaves ********************************************
 *  
 *  - 
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *    - PET cerror code 
 */
#undef __FUNCT__    
#define __FUNCT__ "PromGrid::SendStatesToSlaves" 
int PromGrid::SendStatesToSlaves()
{
  assert(prevgrid_ && factor_ != prevgrid_->factor_); assert( isActive() );
  const int lastfactor = prevgrid_->factor_;
  const int mype = comm_.mype(), tag = prevgrid_->getNewTag();
  MPI_Comm comm =  prevgrid_->gComm_ ;
  int xx, *pb; size_t dsz = sizeof(gfloat)/sizeof(int);
  PromNode *curr;  
  MPI_Status  status; MPI_Request sreq;
  int __sbuf[PROM_BSZ+1]; 
  int *sbuf = ((size_t)__sbuf%sizeof(double)==0) ? __sbuf: __sbuf+1;

  // send results to slaves
  pb = sbuf; assert(((size_t)pb)%dsz==0);
  for( xx = 0 ; xx < nNodes_ ; xx++ ){ 
    curr = nodes_[xx];
    if( curr->getState() == PROM_UNKNOWN ) SETERRQ(1,"st==PROM_UNKNOWN");
    *pb++ = curr->index_.global(); 
    *pb++ = curr->getState();
    PromNode *par = curr->symbData_->parent;
    if( par ) *pb++ = par->index_.global();
    else *pb++ = -1;
    //*pb++ = par ? par->index_.global(): -1;
    if( xx >= nLocalNd_ ) { *pb++ = curr->getNDF(); *pb++ = curr->geq(); }
    if( pb-sbuf >= PROM_BSZ-5 ) {
      // send(tag) *< gid, state, pgid, [ndf, geq] >
      MPI_Isend(sbuf, pb-sbuf,MPI_INT,mype/lastfactor+1,tag, comm, &sreq);
      MPI_Wait( &sreq, &status ); 
      pb = sbuf;
    }
  }
  if( pb-sbuf ) {
    // send(tag) *< gid, state, pgid, [ndf, geq] >
    MPI_Isend(sbuf, pb-sbuf, MPI_INT, mype/lastfactor+1,tag, comm, &sreq);
    MPI_Wait( &sreq, &status ); 
  }
  MPI_Send(NULL, 0, MPI_INT, mype/lastfactor+1, tag, comm);

  return 0;
}  
  
/* PromGrid::ReceiveStates ***********************************************
 *  
 *  - 
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *    - PET cerror code 
 */
#undef __FUNCT__    
#define __FUNCT__ "PromGrid::ReceiveStates" 
int PromGrid::ReceiveStates()
{
  assert(this!=NULL); assert( prevgrid_->isActive() && !isActive() );
  const int mype = comm_.mype(), tag = prevgrid_->getNewTag();
  const int   endp = getEndProc();
  const int   lastfactor = prevgrid_ ? prevgrid_->factor_ : factor_;
  MPI_Comm comm = prevgrid_ ? prevgrid_->gComm_ : gComm_;
  int         xx,*pb,pgid,gid,ierr,done,tt,geq,ndf,*buff;
  MPI_Request sreq; size_t dsz = sizeof(gfloat)/sizeof(int);
  MPI_Status  status;
  PromNode  *curr, *par;
  PromList<int>   newSelected;
  PromList<PromNode*> deleted, extraGhosts;
  PromNodeState state;
  int rbuf[PROM_BSZ]; 

  // receive results
  pb = rbuf; 
  int sproc = (mype+lastfactor), rproc = (mype-lastfactor); 
  xx = 0; 
  do{
    MPI_Probe( MPI_ANY_SOURCE, tag, comm, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    if( tt <= PROM_BSZ ) buff = rbuf;
    else{ ierr = PetscMalloc(tt*sizeof(int), &buff ); CHKERRQ(ierr); }
    pb = buff;
    // recv(tag) *< gid, state, parent, [ndf, geq] >
    MPI_Recv( buff, PROM_BSZ, MPI_INT, rproc/lastfactor, tag, comm, &status );
    pb = buff;  
    assert(((size_t)pb)%dsz==0);
    done = (tt==0); 
    if( sproc < endp ) {
      MPI_Isend( buff, tt, MPI_INT, sproc/lastfactor, tag, comm, &sreq );
    }   
    for( pb = buff ; pb-buff < tt ; xx++ ){ 
      gid = *pb++; state = (PromNodeState)(*pb++); pgid = *pb++;
      if( xx >= nLocalNd_ ) { ndf = *pb++; geq = *pb++; }
      if( xx < nNodes_ ) curr = nodes_[xx];
      else {
	curr = newNode(xx); CHKNEW(curr,1);
	int iii = ghosts_.add(curr, gid+1); assert(iii);
	curr->index_.setGlobal( gid );
	ierr = curr->SetNDF(ndf);  CHKERRQ(ierr); 
	curr->setGEQ(geq);
	extraGhosts.addTail( curr );
      }
      assert(curr!=NULL); assert(curr->index_.global() == gid);
      curr->setState( state );
      if(curr->getState() == PROM_UNKNOWN){SETERRQ(1,"st==PROM_UNKNOWN");}
      else if( curr->getState() == PROM_DELETED ) {
	assert(pgid >= 0);
	int id = pgid - proc_gnode_[masterProc(mype)];
	if( id >= 0 && id < nLocalNd_ ) par = nodes_[id];
	else par = ghosts_.find(pgid+1);
	if( !par ){
	  newSelected.addTail(pgid+1);
	  deleted.addTail(curr);
	}
	else { 
	  assert(!curr->symbData_->parent); 
	  curr->symbData_->parent = par; 
	}
      }
    }
    assert(pb-buff == tt);
    if( sproc < endp ) MPI_Wait( &sreq, &status );
    if( buff != rbuf ) PetscFree(buff);
  }while( !done );
  
  // set parents to news
  Position ipos, pos = deleted.getHeadPosition();
  ipos = newSelected.getHeadPosition();
  while( pos ) {
    curr = deleted.getNext( pos );
    pgid = newSelected.getNext( ipos ) - 1;
    par = ghosts_.find(pgid+1); assert(par!=NULL);
    assert(!curr->symbData_->parent); 
    curr->symbData_->parent = par;
  }
  // add new nodes
  xx = nNodes_; // old value
  ierr = AddNNodePointers( extraGhosts.getCount() ); CHKERRQ(ierr); 
  while( !extraGhosts.isEmpty() ) {
    curr = extraGhosts.removeHead(); assert(!nodes_[xx]); 
    nodes_[xx++] = curr;
  }
  assert(nNodes_ == xx); 
   
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGrid::SquareGraph" 
/* PromGrid::SquareGraph ***********************************************
 *
 *  - G <- G * G
 *
 *  INPUT:
 *   - origadjs: (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc cerror code 
 */
int PromGrid::SquareGraph( int **orig_adj_out )
{
  int ierr,xx,jj,tt,*orig_adj; 
  const int mype = comm_.mype(), tsz = 2024; 
  PromTable<PromNode*> table(229);
  PromNode **curr_arr; int arr_index; PromList<PromNode**> arr_list(1024);
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_AUX3]);
  // 
  for( xx = nLocalNd_ - 1, tt = 0 ; xx >= 0 ; xx-- ) {
    tt += nodes_[xx]->symbData_->adjac.getCount();
  }
  const int nadaj = tt + nLocalNd_;
  ierr = PetscMalloc((nadaj+1)*sizeof(int),&orig_adj); CHKERRQ(ierr);	
  *orig_adj_out = orig_adj;
  orig_adj[nadaj] = -8; if(nadaj>0)orig_adj[nadaj-1] = -8;
  ierr = PetscMalloc(tsz*sizeof(PromNode*),&curr_arr); CHKERRQ(ierr);	
  arr_list.addTail( curr_arr ); arr_index = 0;

  // make extras
  for( xx = tt = jj = 0 ; xx < nLocalNd_ ; xx++ ){
    const PromNode *const curr = nodes_[xx]; 
    const PromTable<PromNode*> &adj0 = curr->symbData_->adjac;
    TablePos tpos = adj0.getHeadPosition();
    while( tpos ){
      PromNode *n1 = adj0.getNext( tpos );
      orig_adj[jj++] = n1->index_();
      if(!curr->isExtended() && !n1->isExtended()){// skip near bound (!parr)
	if( !n1->isExtended() ) {
	  const PromTable<PromNode*> &adj1 = n1->symbData_->adjac;
	  TablePos tpos2 = adj1.getHeadPosition(); 
	  while( tpos2 ){
	    PromNode *n2 = adj1.getNext(tpos2);
	    if( !n2->isExtended() && n2 != curr ) table.add(n2);
	  }
	}
      }
    }
    orig_adj[jj++] = -9;
    //
    while( !table.isEmpty() ) {
      PromNode *n2 = table.removeHead(); 
      if( !adj0.find(n2) ) {
	if( arr_index == tsz ) {
	  ierr = PetscMalloc(tsz*sizeof(PromNode*),&curr_arr); CHKERRQ(ierr);
	  arr_list.addTail( curr_arr ); arr_index = 0;
	}
	curr_arr[arr_index++] = n2;
      }
    }
    //
    if( arr_index == tsz ) {
      ierr = PetscMalloc(tsz*sizeof(PromNode*),&curr_arr); CHKERRQ(ierr);
      arr_list.addTail( curr_arr ); arr_index = 0;
    }
    curr_arr[arr_index++] = NULL;
  }
  assert(tt%2 == 0); assert(nadaj == jj); assert(orig_adj[nadaj] == -8);

  // add these extras
  curr_arr = arr_list.removeHead(); arr_index = 0;
  for( xx = jj = 0 ; xx < nLocalNd_ ; xx++ ) {
    const PromNode *const curr = nodes_[xx];
    PromTable<PromNode*> &adj0 = curr->symbData_->adjac;
    while(1){ 
      PromNode *n2 = curr_arr[arr_index++]; 
      if( arr_index == tsz ) {
	ierr = PetscFree( curr_arr ); CHKERRQ(ierr);
	if( !arr_list.isEmpty() ) {
	  curr_arr = arr_list.removeHead(); arr_index = 0;
	}
	else{ curr_arr = NULL; assert(xx == nLocalNd_-1 && n2 == NULL); }
      }
      if( n2 == NULL ) break;
      else{ int ad = adj0.add( n2 ); assert(ad==1); }
    }
    jj += adj0.getCount();
  }
  assert( arr_list.isEmpty() );
  if( curr_arr != NULL ){ ierr = PetscFree(curr_arr); CHKERRQ(ierr); }
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_AUX3]); 
  if(options_.verbose_>1){
    PetscLogDouble rss; PetscMemoryGetCurrentUsage( &rss );
    PetscPrintf(gComm_,
		"\t[%d]%s two level MIS local edges (ave. |adj|=%d) %d Mb\n",
		mype,__FUNCT__,jj/nLocalNd_,(int)(rss/1.e6+.5));
  }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGrid::UnSquareGraph" 
/* PromGrid::UnSquareGraph ***********************************************
 *  
 *  - Un Do G <- G * G
 *
 *  INPUT:
 *   - orig_adj: (in/out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc cerror code 
 */
int PromGrid::UnSquareGraph( int **orig_adj )
{
  // undo
  int jj, kk, xx, tt, *adjacs = *orig_adj;
  if( adjacs == NULL ) return 0;
  for( xx = 0, tt = 0 , jj = 0; xx < nLocalNd_ ; xx++ ) {
    PromNode *curr = nodes_[xx]; 
    PromTable<PromNode*> &adj = curr->symbData_->adjac;
    adj.removeAll(); 
    while( (kk=adjacs[jj++]) >= 0 ) {
      PromNode *n1 = nodes_[kk];
      int add = adj.add( n1 );  assert(add==1);
    }
    assert(kk==-9);
    tt += adj.getCount();
  }
  assert(adjacs[jj] == -8); assert(jj == 0 || adjacs[jj-1] != -8);
  PetscFree(adjacs);
  if(options_.verbose_>1){
    PetscPrintf(gComm_,"\t[%d]%s undo square grpah ave. |adj|=%d\n",
		comm_.mype(),__FUNCT__,tt/nLocalNd_);
  }
  
  return 0;
}



#undef __FUNCT__    
#define __FUNCT__ "PromGrid::CreateMIS" 
/* PromGrid::CreateMIS ***********************************************
 *
 *  - take this mesh and coarsen, to make (almost) max. ind. sets 
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *    - PETSc cerror code 
 */  
int PromGrid::CreateMIS( PromList<PromNode*> &nodeList, 
			 int Isdeleteable(const PromNode *const,
					  const PromNode *const,bool &),
			 const bool virginMIS /*= FALSE*/ )
{  
  PromTable<int> procTable(57), procTable_orig(57); bool flag;
  int ierr,ii,xx,receive,doit,proc,*pb,ndone,its,todo,todo2,tt,ndep,nbnd;
  const int mype = comm_.mype(),tag=getNewTag(),tag1=getNewTag();
  MPI_Status status;  TablePos tpos;                    
  Position pos,pos2;
  PromNode *curr, *n1;                            assert(isActive());
  TablePos tipos;
  const int nnodes_orig = nNodes_,gtag=getNewTag(),tag2=getNewTag();
  const int reqsz = sizeof(MPI_Request)/sizeof(int);
  PromList<PromNode*> *list;
  PromList<MPI_Request*> smessages; 
  PromTable<PromList<PromNode*>*> proc_list(57);
  PromGlobalID *dep;

  bool determinate = FALSE; 
  ierr = options_.HasOptionName("-prometheus_use_deterministic",&determinate);

  // clear, make gid_ghost and procTable
  for( xx = todo2 = ndep = nbnd = 0 ; xx < nNodes_ ; xx++ ) {
    curr = nodes_[xx];
    curr->setState(PROM_UNKNOWN);    curr->mark(FALSE);
    if( curr->isGhost() ){
      proc = curr->symbData_->ghostID.proc();   procTable_orig.add( proc + 1 );
      if( !proc_list.find(proc+1) ){
	list = new PromList<PromNode*>(100); CHKNEW(list,1);
	proc_list.add( list, proc+1 );
      }
    }
    else todo2++;
    if( curr->isBound() ) {
      nbnd++;
      Position pos = curr->symbData_->dependents->getHeadPosition();
      assert(pos!=NULL);
      while( pos ) {
	dep = curr->symbData_->dependents->getNext(pos); ndep++;
	proc = dep->proc();
	if( !proc_list.find(proc+1) ){
	  list = new PromList<PromNode*>(100); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	  procTable_orig.add( proc + 1 );
	}
      }
    }
  }
  assert(todo2==nLocalNd_);  
  assert(todo2 == nodeList.getCount()); 

  // copy in
  tipos = procTable_orig.getHeadPosition(); 
  while( tipos ) procTable.add( procTable_orig.getNext(tipos) );
  // MIS
  int ptodo = procTable_orig.getCount();
  for( its = 0, todo = 0 ; /* VOID */ ; its++ ){
    // one MIS iteration
    pos = nodeList.getHeadPosition();  ndone = 0;
    while( (pos2=pos) ){
      curr = nodeList.getNext(pos);  
      if( curr->getState() != PROM_UNKNOWN ){ 
	nodeList.removeAt(pos2); 
	// send DELETED bound
	if( curr->getState() == PROM_DELETED && !curr->isMarked() ) {
	  curr->mark();
	  if( curr->isBound() ) {
	    n1 = curr->symbData_->parent; 
	    assert(n1!=NULL && n1->getState()==PROM_SELECTED);
	    // send to deps
	    Position pos=curr->symbData_->dependents->getHeadPosition();
	    assert(pos!=NULL);
	    while( pos ) {
	      dep = curr->symbData_->dependents->getNext(pos);
	      proc=dep->proc();
	      list = proc_list.find(proc+1);	      assert(list!=NULL);
	      list->addTail( curr ); list->addTail( n1 ); 
	    }
	  }
	} 
	continue; 
      } 
      // check for selectable
      doit = TRUE; //(!curr->isAllFixed() || its>0); // ***
      tpos = curr->symbData_->adjac.getHeadPosition();  
      while( tpos && doit ) {
	n1 = curr->symbData_->adjac.getNext( tpos );
	//if( n1->isAllFixed() ) continue; // ***
	if( n1->getState() == PROM_UNKNOWN ) { 
	  int pp = n1->isGhost() ? n1->symbData_->ghostID.proc() : mype;
	  if( (curr->getTopoType()==n1->getTopoType() && pp > mype) ||
	      curr->getTopoType() > n1->getTopoType() ) { 
	    ierr = Isdeleteable( n1, curr, flag ); CHKERRQ(ierr); 
	    if( flag ) doit = FALSE;
	  }
	}
	else if( n1->getState() == PROM_SELECTED ) {
	  ierr = Isdeleteable( n1, curr, flag ); CHKERRQ(ierr); 
	  if( flag ) {
	    doit = FALSE; assert( n1->isGhost() );  
	  }
	}
	else if( virginMIS && its == 0 ) {
	  assert(n1->getState()==PROM_DELETED);
	  doit = FALSE; 
	}
      }
      if( doit ) {
	// select
	curr->setState(PROM_SELECTED);  todo2--; 
	tpos = curr->symbData_->adjac.getHeadPosition();   
	while( tpos ) { 
	  n1 = curr->symbData_->adjac.getNext( tpos );
	  if( n1->getState() == PROM_UNKNOWN ) {
	    ierr = Isdeleteable( curr, n1, flag ); CHKERRQ(ierr);
	    if( flag /*|| n1->isAllFixed()*/ ) { // ***
	      if( n1->isGhost() ){
		proc = n1->symbData_->ghostID.proc();  assert(proc!=mype);
		list = proc_list.find(proc+1); assert(list!=NULL);
		list->addTail( n1 ); list->addTail( curr );  
	      }
	      else todo2--; 
	      n1->setState( PROM_DELETED );  
	      n1->symbData_->parent = curr;
	      ndone++;
	    }
	  }
	  else if( n1->getState() == PROM_DELETED ) {
	    if( !n1->symbData_->parent->isGhost() ) ii = mype;
	    else ii = n1->symbData_->parent->symbData_->ghostID.proc();
	    if( ii < mype ) {
	      ierr = Isdeleteable( curr, n1, flag ); CHKERRQ(ierr);
	      if( flag /*|| n1->isAllFixed()*/ ){ // ***
		n1->symbData_->parent = curr; 
		if( n1->isGhost() ){
		  proc = n1->symbData_->ghostID.proc();  assert(proc!=mype);
		  list = proc_list.find(proc+1); assert(list!=NULL);
		  list->addTail( n1 ); list->addTail( curr );
		}
		else n1->mark(FALSE);
	      }
	    }
	  } 
	}
      }
    }

    // isend(tag)
    tpos = proc_list.getHeadPosition();
    while( tpos ){ 
      list = proc_list.getNext( tpos, &proc ); proc--;
      if( list->isEmpty() ) continue; // do not send NUL as is capper flag
      tt = list->getCount();                      assert(tt%2==0 && tt>0);
      if( tt ) {
	MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
	CHKERRQ(ierr);	smessages.addTail( sreq );
	int *buff = (int*)(sreq+1), *pb = buff;
	while( !list->isEmpty() ) { 
	  n1 = list->removeHead(); *pb++ = n1->index_.global();
	}
	assert(tt == pb-buff);
	// send(tag) *<gid_D_gh, gid_S_loc> | <gid_loc_DEL, pargid_SEL>
	MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag, gComm_, sreq);
      }
    }
    // loop test
    if( determinate ) {
      MPI_Allreduce( &todo2, &tt, 1, MPI_INT, MPI_MAX, gComm_ );
      if ( tt == 0 ) break;
    }
    else if ( todo2 == 0 ) break;
    
    // receive(tag)
    do{
      tpos = proc_list.getHeadPosition();  
      while( tpos ) { 
	proc_list.getNext( tpos, &proc ); proc--;
	do{
	  MPI_Iprobe( proc/factor_, tag, gComm_, &receive, &status );
	  if( receive ) {
	    MPI_Get_count( &status, MPI_INT, &tt ); 
	    if( tt == 0 ) {
	      MPI_Recv(NULL,0,MPI_INT,status.MPI_SOURCE,tag,gComm_,&status);
	      ptodo--;
	    }
	    else {
	      // recv(tag) *< gid_D_loc, gid_S_gh> | < gid_gh_D, pargid_S>
	      ierr = ReceiveState_NC( ghosts_, tt, ndone, todo2, todo, tag,
				      gtag, status.MPI_SOURCE, gComm_ );
	      CHKERRQ(ierr);
	    }
	  }
	}while( receive && todo2 );
      }
    }while( ndone == 0 && todo2 ); // wait till something to do
  } // main MIS loop

  // collect PROM_DELETED bound
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ) {
    curr = nodes_[xx];
    if( curr->getState() == PROM_DELETED && !curr->isMarked() ) {
      curr->mark();
      if( curr->isBound() ) {
	n1 = curr->symbData_->parent;  
	assert(n1!=NULL && n1->getState()==PROM_SELECTED);
	Position pos = curr->symbData_->dependents->getHeadPosition();
	assert(pos!=NULL);
	while( pos ) {
	  dep = curr->symbData_->dependents->getNext( pos ); 
	  proc = dep->proc();
	  list = proc_list.find(proc+1); assert(list!=NULL);
	  list->addTail( curr ); list->addTail( n1 ); 
	}
      }
    }
  }
  // send PROM_DELETED bound
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc); proc--;
    tt = list->getCount(); 
    if( tt ) {
      MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
      CHKERRQ(ierr);	smessages.addTail( sreq );
      int *buff = (int*)(sreq+1), *pb = buff;
      while( !list->isEmpty() ) {
	curr = list->removeHead(); 
	*pb++ = curr->index_.global();
      }
      assert(tt == pb-buff);
      // send(tag) *<gid_D_loc, gid_S_gh> | <gid_gh_D, pargid_S>
      MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag, gComm_, sreq );

    }
    delete list; 
  }

  // send(tag) capper
  tipos = procTable_orig.getHeadPosition();
  while( tipos ) {
    proc = procTable_orig.getNext(tipos) - 1;
    MPI_Send( NULL, 0, MPI_INT, proc/factor_, tag, gComm_ );
  }
  // receive(tag) and capper(tag) 
  do{
    MPI_Iprobe( MPI_ANY_SOURCE, tag, gComm_, &receive, &status );
    if( receive ) {
      MPI_Get_count( &status, MPI_INT, &tt ); 
      if( tt == 0 ) {
	MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status);
	ptodo--; assert(ptodo>=0);
      }
      else {
	// recv(tag) *< gid_D_loc, gid_S_gh> | < gid_gh_D, pargid_S >
	ierr = ReceiveState_NC( ghosts_, tt, ndone, todo2, todo, tag, gtag,
				status.MPI_SOURCE, gComm_ );
	CHKERRQ(ierr);
      }
    }
  }while( ptodo );

  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }

  // collect send(tag2) PROM_DELETED bound
  assert( proc_list.isEmpty() );
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ) {
    curr = nodes_[xx];  
    if( curr->getState() == PROM_DELETED && !curr->isMarked() ) {
      curr->mark();
      if( curr->isBound() ) {
	// collect send to deps
	Position pos = curr->symbData_->dependents->getHeadPosition();
	while( pos ) {
	  dep = curr->symbData_->dependents->getNext( pos );
	  proc = dep->proc();
	  if( (list=proc_list.find(proc+1)) == NULL ){
	    list = new PromList<PromNode*>(10); CHKNEW(list,1);
	    proc_list.add( list, proc+1 );
	  }
	  list->addTail( curr );
	}
      }
    }
  }

  // send(tag2) PROM_DELETED bound
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc); proc--;
    tt = 2*list->getCount(); 
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);	smessages.addTail( sreq );
    int *buff = (int*)(sreq+1), *pb = buff;
    while( !list->isEmpty() ) {
      curr = list->removeHead();
      n1 = curr->symbData_->parent;
      *pb++ = curr->index_.global(); *pb++ = n1->index_.global();
    }
    assert(tt == pb-buff);
    // send(tag2) *< gid_loc_DEL, pargid_SEL >
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag2, gComm_, sreq );
    delete list;
  }

  // capper (tag2)
  tipos = procTable_orig.getHeadPosition(); ptodo = 0;
  while( tipos ) {
    proc = procTable_orig.getNext(tipos) - 1; ptodo++;     
    MPI_Send(NULL, 0, MPI_INT, proc/factor_, tag2, gComm_ );
  }
  while( ptodo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag2, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    if( tt == 0 ) {
      ptodo--;
      MPI_Recv(NULL,0,MPI_INT,status.MPI_SOURCE,tag2,gComm_,&status);
    }
    else {
      // recv(tag2) *< gid_loc_DEL, pargid_SEL >
      ierr = ReceiveState_NC( ghosts_, tt, ndone, todo2, todo, tag2, gtag,
			      status.MPI_SOURCE, gComm_ );
      CHKERRQ(ierr);
    } 
  }

  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }

  // set extras selected (will get done in add...)
  for( xx = nLocalNd_ ; xx < nnodes_orig ; xx++ ) {
    curr = nodes_[xx];
    if( curr->getState() == PROM_UNKNOWN ) curr->setState( PROM_SELECTED );
  }

  { // get procs of neigbors
    int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 
    pb = sbuf;
    tipos = procTable_orig.getHeadPosition(); 
    while( tipos ) *pb++ = procTable_orig.getNext(tipos) - 1; 
    int nsen = procTable_orig.getCount(); 
    MPI_Request *sreqs = NULL; MPI_Status *statuses = NULL;
    if( nsen ) {
      ierr = PetscMalloc( sizeof(MPI_Request)*nsen, &sreqs );    CHKERRQ(ierr);
      ierr = PetscMalloc( sizeof(MPI_Status)*nsen, &statuses );  CHKERRQ(ierr);
    }
    // send(tag1) <orig procs>
    for( tipos = procTable_orig.getHeadPosition(), ii = 0 ; tipos ; ii++ ) {
      proc = procTable_orig.getNext(tipos) - 1;     
      MPI_Isend(sbuf,pb-sbuf,MPI_INT, proc/factor_,tag1,gComm_,&sreqs[ii]);
    }
    // receive(tag1)
    tipos = procTable_orig.getHeadPosition(); 
    while( tipos ) {
      procTable_orig.getNext(tipos); // counter only
      MPI_Recv( rbuf, PROM_BSZ, MPI_INT, MPI_ANY_SOURCE, tag1,gComm_,&status);
      MPI_Get_count( &status, MPI_INT, &tt );
      for ( pb = rbuf ; tt ; tt-- ){
	proc = *pb++; if( proc != mype ) procTable.add( proc+1 );
      }
    }
    if( nsen ) {
      MPI_Waitall(nsen,sreqs,statuses); PetscFree(sreqs); PetscFree(statuses);
    }
  }
  if(options_.verbose_>1)
    PetscPrintf(gComm_,"[%d]%s %d new ghosts(1)\n",mype,__FUNCT__,todo);
  // set data structures
  ierr = ReceiveAddGhost( todo, gtag, procTable ); CHKERRQ(ierr);
  
  // clean up and check
  for( xx = 0 ; xx < nNodes_ ; xx++ ) nodes_[xx]->mark(FALSE);
  MPI_Iprobe(MPI_ANY_SOURCE, tag, gComm_, &receive, &status);assert(!receive);
  MPI_Iprobe(MPI_ANY_SOURCE, tag1, gComm_, &receive,&status);assert(!receive);
  MPI_Iprobe(MPI_ANY_SOURCE, tag2, gComm_, &receive,&status);assert(!receive);
  MPI_Iprobe(MPI_ANY_SOURCE, gtag, gComm_, &receive,&status);assert(!receive);

  return 0;
}


#undef __FUNCT__  
#define __FUNCT__ "PromGrid::ReceiveState_NC" 
/* PromGrid::ReceiveState_NC ********************************************
 *
 *  INPUT:
 *   - gid_ghost:
 *   - sz: 
 *   - ndone: (in/out)
 *   - todo2: (in/out)
 *   - todo: (in/out)
 *   - tag: this and proc could just be the status ...
 *   - part:
 *   - comm: 
 *
 *  SIDE EFFECTS:
 *  RETURN: 
 *     - PETSc error code 
 */
int PromGrid::ReceiveState_NC( PromTable<PromNode*> &gid_ghost, const int sz, 
			       int &ndone, int &todo2, int &todo, 
			       const int rtag, const int stag, const int part,
			       MPI_Comm comm )
{
  int ii, gid, gid2, proc2, ierr, *buff, *pb;
  const int mype = comm_.mype();
  MPI_Status  status;
  PromNode  *select,*del; 
  int rbuf[PROM_BSZ]; 

  if( sz <= PROM_BSZ ) buff = rbuf;
  else{ ierr = PetscMalloc(sz*sizeof(int),&buff); CHKERRQ(ierr); }
  pb = buff; assert(sz%2==0 && sz>0); 
  
  // recv(rtag) < gid_D_loc, gid_S_gh > or < gid_D_gh, par_S >
  MPI_Recv( buff, sz, MPI_INT, part, rtag, comm, &status ); 
  while( pb-buff < sz ) {
    gid = *pb++;  gid2 = *pb++; 
    del = gid_ghost.find( gid+1 ); 
    if( del == NULL ) { // deleted local
      // recv(rtag) < gid_DEL_loc, gid2_SEL_gh >
      select = gid_ghost.find( gid2+1 ); assert(select!=NULL);
      // gid_PROM_DELETED_loc
      ii = gid - proc_gnode_[mype]; assert(ii>=0 && ii < nLocalNd_);
      del = nodes_[ii];      assert(del->getState() != PROM_SELECTED); 
      if( select->getState() == PROM_UNKNOWN ) {
	select->setState(PROM_SELECTED);        	ndone++; 
      }
      if( del->getState() == PROM_UNKNOWN ){
	del->setState(PROM_DELETED); del->symbData_->parent = select; todo2--; 
      } 
      else {
	assert(del->getState() == PROM_DELETED && del->symbData_->parent);
	if( !del->symbData_->parent->isGhost() ) proc2 = mype;
	else proc2 = del->symbData_->parent->symbData_->ghostID.proc();
	if( select->symbData_->ghostID.proc() > proc2 ){
	  del->symbData_->parent = select;  // PROM_DELETED
	  del->mark(FALSE); // send again as changed
	}
      }
    }
    else { // delete ghost
      // recv(rtag) select < gid_D_gh, pargid_S > 
      ii = gid2 - proc_gnode_[mype]; 
      if( ii >= 0 && ii < nLocalNd_ ){ select = nodes_[ii]; proc2 = mype; }
      else {
	ierr = GetProc( gid2, proc2 );  CHKERRQ(ierr); assert(proc2!=mype);
	todo += addGhost(select, proc2, gid2, gid2 - proc_gnode_[proc2], stag);
      }
      assert(select->getState() != PROM_DELETED);
      select->setState( PROM_SELECTED ); 
      if( del->getState() == PROM_UNKNOWN ) {  
	del->setState(PROM_DELETED); del->symbData_->parent = select;  ndone++;
      }
      else{
	assert(del->getState() == PROM_DELETED);
	assert(del->symbData_->parent != NULL);
	int pp = del->symbData_->parent->isGhost() ?
	  del->symbData_->parent->symbData_->ghostID.proc() : mype;
	if( proc2 == mype ) assert(del->symbData_->parent == select);
	if( proc2 > pp ) del->symbData_->parent = select;  // DELETE -- change
      }
    }
  }
  assert(pb-buff==sz);
  if( buff != rbuf )  PetscFree(buff);
  
  return 0;
}
